from ws4py.client.threadedclient import WebSocketClient as _WebSocketClient

import ssl
import logging


logging.basicConfig(level=logging.INFO,
                    filename='/opt/amqp-middleware/agent.log',
                    format='%(asctime)s [%(levelname)s] - '
                           '%(funcName)s: %(message)s')

log = logging.getLogger(__name__)


class WebSocketClient(_WebSocketClient):
    """A simple WebSocket client."""

    def __init__(self, url, token, secure=False, ca_certs=None, validate=True):
        """Initializes a WebSocket client and connects to the server.

        :param url: the address of the server.
        :param token: the token to be used for authentication with the server.
        :param ca_certs: the file of CA certificates to be used in case of SSL.
        :param validate: specifies whether a certificate is required.
        :param secure: specifies whether or not to use a secure ws.

        """
        scheme = 'ws'
        ssl_options = {}
        if secure:
            if not ca_certs:
                raise Exception('SSL is enabled, but path to ca_certs is '
                                'missing')
            scheme = 'wss'
            reqs = ssl.CERT_REQUIRED if validate else ssl.CERT_NONE
            ssl_options = {
                'ca_certs': ca_certs,
                'cert_reqs': reqs
            }

        super(WebSocketClient, self).__init__(
            '%s://%s/insights/socket/' % (scheme, url),
            headers=[('Authorization', token)], ssl_options=ssl_options
        )

        self.connect()

    def opened(self):
        """Invoked when a new WebSocket is opened."""
        log.info('Connection OPEN')

    def closed(self, code, reason):
        """Invoked when the WebSocket is closed."""
        log.warning('Connection CLOSED with code %s: %s', code, reason)

    def write_message(self, data):
        """Sends the given payload out."""
        self.send(data)



import unittest

import pika
import json
import time
import requests
import threading

import amqpconsumer.components.agent


class AMQPTest(unittest.TestCase):

    def setUp(self):
        reg = requests.post('http://localhost/api/v1/insights/register',
                            data=json.dumps({'name': 'Unit Tester',
                                             'email': 'tester@unit.io'})
                           )
        if not reg.ok:
            raise Exception('Registration failed')

        res = reg.json()
        self.id = res['uuid']

        self.ws_kwargs = {
            url='localhost:8888/insights/socket',
            token=res['token'],
            secure=False,
            ca_certs=None,
        }

        self.cfy_kwargs = {
            'host': 'localhost',  # The API call to Cloudify Manager will fail.
            'secure': False,
            'verify': True,
            'credentials': {
                'username': '',
                'password': '',
                'ca_certs': None,
            },
            'ssl_enabled': False,
        }

        self.rmq_kwargs = {
            'host': 'localhost',
            'port': 5672,
            'credentials': {
                'username': 'guest',
                'password': 'guest',
                'ca_certs': None,
                'ssl_enabled': False,
             },
        }

    def tearDown(self):
        pass

    def test(self):
        publisher = amqpconsumer.components.agent.Publisher(
            id=self.id,
            kwargs={'ws': self.ws_kwargs, 'cfy': self.cfy_kwargs}
        )

        consumer = amqpconsumer.components.agent.AMQPConsumer(
            exchange='cloudify-metrics',
            exchange_type='topic', routing_key='*',
            data_processor=publisher.process_event,
            **self.rmq_kwargs,
        )

        def start(agent):
            agent.start()

        thread = threading.Thread(target=start, kwargs={'agent': consumer})
        thread.daemon = True
        thread.start()
        time.sleep(2)
        publish_event()
        thread.join(5)

        self.assertTrue(find_event(), 'Failed to find ES document')


def publish_event():

    event = {
        'host': 'dummy_host',
        'node_id': 'dummy_id',
        'node_name': 'dummy_name',
        'deployment_id': 'dummy_deployment',
        'name': 'test',
        'path': 'path',
        'metric': 1,
        'unit': '',
        'type': 'test_type',
        'time': time.time(),
    }

    connection = pika.BlockingConnection(
        pika.ConnectionParameters(host='localhost')
    )
    channel = connection.channel()
    channel.exchange_declare(exchange='cloudify-metrics', type='topic',
                             durable=False, auto_delete=True, internal=False)
    channel.basic_publish(exchange='cloudify-metrics', routing_key='metrics',
                          body=json.dumps(event))
    channel.close()
    connection.close()


def find_event():
    query = {
        "query": {
            "bool": {
                "must": [
                    {"match": {"@fields.host": "dummy_host"}},
                    {"match": {"@fields.node_id": "dummy_id"}},
                    {"match": {"@fields.node_name": "dummy_name"}},
                    {"match": {"@fields.deployment_id": "dummy_deployment"}},
                ],
            },
        },
    }

    req = requests.get('http://localhost:9200/*insights*/_search',
                        data=json.dumps(query))
    if req.ok:
        hits = req.json()['hits']['hits']
        if hits:
            return True
    return False


if __name__ == '__main__':
    unittest.main()



import logging

from requests import Session


logging.getLogger('requests').setLevel(logging.WARNING)
logging.basicConfig(level=logging.INFO,
                    filename='/opt/amqp-middleware/agent.log',
                    format='%(asctime)s [%(levelname)s] - '
                           '%(funcName)s: %(message)s')

log = logging.getLogger(__name__)

# A mapping between cloud provider names and the field of the corresponding
# resource ID, as it is meant to exist in the database.
PROVIDERS = {
    'azure': 'name',
    'do': 'resource_id',
    'aws': 'aws_resource_id',
    'docker': 'container_id',
    'openstack': 'external_id',
    'vcloud': 'vcloud_vapp_name',
    'vsphere': 'vsphere_server_id',
}


class CloudifyClient(object):
    """An HTTP client for Cloudify Manager's API."""

    def __init__(self, host, credentials, secure=False, ssl_enabled=False,
                 verify=True):
        """The CloudifyClient queries the Cloudify Manager's API in order to
        retrieve information on deployments and node instances.

        :param host: the Cloudify Manager's IP address.
        :param credentials: security parameters used in order to authenticate
         to a secure Cloudify Manager.
        :param secure: denotes whether the Cloudify Manager is secure.
        :param ssl_enabled: denotes whether SSL is enabled.
        :param verify: specifies whether to verify the SSL certificate.

        """
        username = credentials.get('username', '')
        password = credentials.get('password', '')
        tenant = credentials.get('tenant', '')
        if secure and not (username and password):
            raise Exception('Security is enabled, but no username:password '
                            'has been provided')
        if not tenant:
            raise Exception('No tenant specified')
        if ssl_enabled:
            cert = credentials.get('ca_certs', '')
            if not cert:
                raise Exception('SSL enabled, but path to public certificate '
                                'is missing')

        scheme = 'https' if (secure and ssl_enabled) else 'http'
        self.url = '%s://%s/api/v2.1/node-instances' % (scheme, host)
        self.session = Session()
        self.session.headers.update({'Tenant': tenant})
        if secure:
            self.session.auth = (username, password)
        if ssl_enabled:
            if secure:
                self.session.verify = cert if verify else verify
            else:
                # https://github.com/cloudify-cosmo/
                # cloudify-manager-blueprints/blob/3.4-build/
                # aws-ec2-manager-blueprint-inputs.yaml#L58
                log.warning('Ignoring SSL, since security is disabled')

        # Cached host information in the form of:
        # { host_id: ( provider, resource_id ) }
        self.resources = dict()

    def get_rid(self, event):
        """Maps the host ID (assigned by Cloudify Manager) contained in the
        event dict to the correponding cloud provider's resource ID.
        """
        deployment_id = event['deployment_id']
        host_id = event['host']

        if host_id not in self.resources:
            log.info('No cached information found for host "%s" (%s)',
                     host_id, deployment_id)

            # Query db for the node instance's runtime properties.
            runtime_properties = self.get_attributes(deployment_id, host_id)

            # Search for host information and cache it.
            for provider, rfield in PROVIDERS.iteritems():
                if rfield in runtime_properties.iterkeys():
                    rid = runtime_properties[rfield]

                    self.resources[host_id] = (provider, rid)
                    log.info('Database HIT for host "%s" (%s): %s = %s',
                             host_id, deployment_id, rfield, rid)
                    break
            else:
                log.error('Failed to find resource ID for host "%s" (%s)',
                          host_id, deployment_id)
                return

        # Inject resource ID in the event.
        provider, rid = self.resources[host_id]
        event.update({
            'provider': provider,
            'resource_id': rid
        })

    def get_attributes(self, deployment_id, host_id):
        """Queries the Cloudify Manager's API for the specified `host_id` and
        verifies that the deployment the node instance belongs to matches the
        `deployment_id` provided.

        If the query fails, this method will not raise an error, but rather
        return. In that case, the provider's resource ID will NOT be injected
        into the metric, which may cause identification issues later on.

        :param deployment_id: the deployment the node instance belongs to.
        :param host_id: the node instance's host ID.

        :return: the node instance's runtime properties.

        """
        node_instance = self.session.get('%s/%s' % (self.url, host_id))

        try:
            node_instance = node_instance.json()
            assert node_instance['deployment_id'] == deployment_id
        except ValueError as err:
            log.error('HTTP request failed for node "%s" (%s) with status '
                      'code %d: %s', host_id, deployment_id,
                      node_instance.status_code, node_instance.content)
            # Propagate the original error in case the API returned OK,
            # but decoding failed regardless.
            if node_instance.ok:
                raise
        except AssertionError as err:
            log.error('Deployment ID mismatch for host "%s" (%s)',
                      host_id, deployment_id)
        else:
            err = ''

        return node_instance['runtime_properties'] if not err else {}



import os
import ssl
import pika
import json
import time
import logging

from amqpconsumer.components.manager import CloudifyClient
from amqpconsumer.components.connection import WebSocketClient


logging.getLogger('requests').setLevel(logging.WARNING)
logging.basicConfig(level=logging.INFO,
                    filename='/opt/amqp-middleware/agent.log',
                    format='%(asctime)s [%(levelname)s] - '
                           '%(funcName)s: %(message)s')

log = logging.getLogger(__name__)


class AMQPConsumer(object):
    """An AMQP Middleware able to consume data from RabbitMQ and publish data
    to the desired destination endpoint through a WebSocket.
    """

    def __init__(self, host, port, credentials, data_processor,
                 exchange='', exchange_type='topic', routing_key='*'):
        """Initialize the AMQP Middleware by subscribing to the specified
        RabbitMQ exchange.

        :param host: the IP address of the RabbitMQ server.
        :param port: the port the RMQ server is listening at.
        :param credentials: credentials required to establish the connection.
        :param data_processor: invoked by the AMQPConsumer's callback method
         upon a receiving a new message.
        :param exchange: the RMQ exchange to subscribe to.
        :param exchange_type: the type of the RMQ exchange.
        :param routing_key: the routing key to be used.

        """
        self.process_data = data_processor

        connection_params = {'host': host, 'port': int(port)}
        connection_params = self.update_params(connection_params, credentials)

        # The `pika.adapters.blocking_connection.BlockingConnection` does not
        # respect the `connection_attempts` parameter.
        # See: pika.readthedocs.io/en/latest/examples/using_urlparameters.html
        retry_attempts = connection_params.pop('connection_attempts')
        retry_delay = connection_params.pop('retry_delay')

        # Wrap in a try/except statement to ensure compatibility with older
        # pika versions, which ignore certain connection parameters.
        # See: https://github.com/pika/pika/issues/354
        for _ in xrange(retry_attempts):
            try:
                self.connection = pika.BlockingConnection(
                    pika.ConnectionParameters(**connection_params)
                )
            except pika.exceptions.AMQPConnectionError:
                time.sleep(retry_delay)
            else:
                self.channel = self.connection.channel()
                break
        else:
            raise pika.exceptions.AMQPConnectionError()

        # Declare RMQ exchange.
        self.channel.exchange_declare(exchange=exchange,
                                      exchange_type=exchange_type,
                                      durable=False, auto_delete=True,
                                      internal=False)

        # Declare and bind queue.
        result = self.channel.queue_declare(durable=False,
                                            exclusive=False,
                                            auto_delete=True)
        queue = result.method.queue
        self.channel.queue_bind(exchange=exchange, queue=queue,
                                routing_key=routing_key)

        self.channel.basic_consume(consumer_callback=self._callback,
                                   queue=queue, no_ack=True)

    @staticmethod
    def update_params(params, creds, max_retries=3, retry_delay=1,
                      socket_timeout=2):
        """Prepares the connection parameters to be used for connecting
        to RabbitMQ.

        :param params: the initial dict of connection parameters.
        :param creds: a dict containing authentication credentials.
        :param max_retries: the number of total connection attempts.
        :param retry_delay: the delay in between connection attempts.
        :param socket_timeout: maximum socket time-out.

        :return: the updated dict of connection parameters to be used.

        """
        params['credentials'] = pika.credentials.PlainCredentials(
            creds['username'], creds['password'])

        if creds['ssl_enabled']:
            if not creds['ca_certs']:
                raise Exception('Missing path to ca_certs')
            params['ssl'] = creds['ssl_enabled']
            params['ssl_options'] = {
                'cert_reqs': ssl.CERT_REQUIRED,
                'ca_certs': creds['ca_certs']
            }

        params['connection_attempts'] = max_retries
        params['retry_delay'] = retry_delay
        params['socket_timeout'] = socket_timeout

        return params

    def start(self):
        """Start consuming data from RabbitMQ."""
        self.channel.start_consuming()

    def _callback(self, channel, method, properties, body):
        """The callback method invoked upon receiving a new message."""
        try:
            self.process_data(json.loads(body))
        except Exception as exc:
            log.error('Data processing failed: %s', exc)
            raise exc


class Publisher(WebSocketClient):
    """The publishing part of the AMQP Middleware."""

    def __init__(self, id, kwargs, max_batch_size=100, max_batch_delay=2):
        """Initializes a WebSocket-based Publisher.

        This part of the AMQP Middleware is responsible for processing new
        messages consumed by the AMQPConsumer and sending them to the
        specified  destination URL in batches.

        The AMQPConsumer directly invokes the Publisher's `process_event`
        method as its callback function. The Publiser also utilizes the
        CloudifyClient in order to query the Cloudify Manager's REST API
        regarding deployments and node instances.

        :param url: the destination URL the data will be sent to.
        :param id: a unique ID among Cloudify Manager installations.
        :param token: a token to be used for authentication on WebSocket open.
        :param secure: specifies whether data will be transmitted over SSL.
        :param ca_certs: the path to the bundle of CA certificates.
        :param validate: denotes whether certificates are required from the
         other side of the connection.
        :param max_batch_size: the size of each batch to send.
        :param max_batch_delay: the maximum period of time to wait between
         consecutive pushes of data.

        """
        super(Publisher, self).__init__(**kwargs.pop('ws'))

        self.cfy = CloudifyClient(**kwargs.pop('cfy'))

        # A UUID associated with the AMQP Middleware.
        self.mid = id

        # Declare maximum batch size/delay.
        self.batch_size = max_batch_size
        self.batch_delay = max_batch_delay

        # Initialize conditions for publishing data.
        self.current_batch = ''
        self.current_batch_size = 0
        self.last_push = time.time()

    def process_event(self, event):
        """Processes new messages received by the AMQPConsumer.

        Publishes batches of data to the desired destination.

        This method is practically the callback function of the AMQPConsumer.
        """
        if 'metric' in event:
            # Process event, before sending it out.
            self._process_event__edit_fields(event)

            # Incrementally create the batch to send out, delimited by `\n`.
            self.current_batch += json.dumps(event) + '\n'
            self.current_batch_size += 1

            # Push data via the WebSocket, if appropriate.
            if self.current_batch_size < self.batch_size and \
               time.time() < self.last_push + self.batch_delay:
                return
            self.send_data(self.current_batch)

            # Reset batch counters.
            self.current_batch = ''
            self.current_batch_size = 0
            self.last_push = time.time()

    def _process_event__edit_fields(self, event):
        """Edits the fields of a new event.

        This method is responsible for injecting into the event dict the
        corresponding cloud provider's resource ID for identification purposes.

        This method is meant to be called by `process_event`.
        """
        event['time'] = time.time()
        event['owner_id'] = str(self.mid)

        # Get the resource ID corresponding to this event.
        self.cfy.get_rid(event)

    def send_data(self, data):
        """Push data through the WebSocket."""
        log.debug('Sending batch of size %s within %s seconds',
                  self.current_batch_size, time.time() - self.last_push)
        self.write_message(json.dumps(data))


def bootstrap():
    """Initialize the AMQP Middleware.

    Main method to start the AMQP Middleware by first fetching the required
    configuration options from the systemd env file.

    NOTE: All boolean expresseions are handled as strings for compatibility
    with Cloudify Blueprints.

    """

    # Set of parameters required by the WebSocket client.
    ws_kwargs = {
        'url': os.getenv('DESTINATION_URL'),
        'token': os.getenv('TOKEN'),
        'secure': True if os.getenv('STREAM_SECURE') == 'True' else False,
        'ca_certs': os.getenv('STREAM_CA_CERTS'),
    }

    # Parameters required to communicate with Cloudify Manager's HTTP API.
    cfy_kwargs = {
        'host': os.getenv('HOST'),
        'secure': True if os.getenv('SECURE') == 'True' else False,
        'verify': True if os.getenv('VERIFY') == 'True' else False,
        'credentials': {
            'tenant': os.getenv('TENANT'),
            'username': os.getenv('USERNAME'),
            'password': os.getenv('PASSWORD'),
            'ca_certs': os.getenv('CA_CERTS'),
        },
        'ssl_enabled': True if os.getenv('SSL_ENABLED') == 'True' else False,
    }

    publisher = Publisher(
        id=os.getenv('MANAGER_ID'),
        kwargs={'ws': ws_kwargs, 'cfy': cfy_kwargs}
    )

    consumer = AMQPConsumer(
        host=os.getenv('RABBITMQ_HOST'),
        port=os.getenv('RABBITMQ_PORT'),
        credentials={
            'username': os.getenv('RABBITMQ_USERNAME'),
            'password': os.getenv('RABBITMQ_PASSWORD'),
            'ca_certs': os.getenv('RABBITMQ_CERT_PUBLIC'),
            'ssl_enabled':
                True if os.getenv('RABBITMQ_SSL_ENABLED') == 'True' else False,
        },
        exchange='cloudify-monitoring',
        exchange_type='topic', routing_key='*',
        data_processor=publisher.process_event
    )

    # Actually start the AMQP Middleware.
    consumer.start()


if __name__ == '__main__':
    bootstrap()



from setuptools import setup, find_packages


setup(
    name='amqp-middleware',
    version='1.1.2',
    description='AMQP Consumer-Publisher Middleware',
    url='https://github.com/mistio/amqp-middleware',
    license='Apache',
    author='Chris Pollalis',
    author_email='cpollalis@mist.io',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'amqp-middleware = amqpconsumer.components.agent:bootstrap',
        ],
    },
    dependency_links=[
        'git+https://github.com/mistio/amqp-middleware.git@master#egg=amqp-middleware',
    ],
    install_requires=[
        'pika>=0.9.14',
        'ws4py==0.3.5',
        'requests==2.7.0',
    ],
)
